//Components
import Utils  from "../API/IKUtils";
import IKDBDDB from "../Database/DDB/IKDB";
import IKDBSQL from "../Database/SQL/IKDB";
import IKLogger from "../Logger/IKLogger";
import IKGlobals from '../IKGlobals';
import IKValidation from '../Validation/IKValidation';
import IKRequest from '../API/IKRequest';
import IKResponse, { IKBadRequestResponseWithRollback } from '../API/IKResponse';
import IKPublisher from '../Publisher/IKPublisher';
//
export default class IKTransaction {
  constructor(event, context, config, _isQueue) {
    //
    this._event = event;
    this._context = context;
    this._config = config;
    //queue support
    this._onQueue = _isQueue;
    this._queueResp = null;
    //
    this.logger = new IKLogger({/*COFIG S3/SQS HERE*/}, Utils.logLevel(), (context.awsRequestId ? context.awsRequestId : (event.requestContext ? event.requestContext.requestId : 'unknown')));
    this.request = new IKRequest(this._event, this._context);
    this.publisher = new IKPublisher(config.publisher);
    this.validator = new IKValidation(config.validation);
    this.db = this._getDBDriver();
  }

  //Main interface
  async execute(executionFunc) {
    await this._executeLoggerFlush(async ()=>{
      await this._executeDBTransactions(async ()=>{
        return await this._execute(executionFunc);
      });
    });
    if (this._onQueue) return this._queueResp;
  }
  //Executions
  async _execute(executionFunc) {
    let executionFailed = true; //failled til we say no!
    //safe execution handler
    try { //Execute
      console.debug("Starting main request code");
      this._queueResp = await executionFunc(this);
      //Answer client
      if (this._queueResp && this._queueResp instanceof IKResponse) {
        this._queueResp.build(this._context, this.request.getRequestID(), this._onQueue);
        executionFailed = false;
      } else {
        this._queueResp = IKBadRequestResponseWithRollback(IKGlobals.ErrorResponseInvalidServerResponse, IKGlobals.ErrorCode_APIError)
        this._queueResp.build(this._context, this.request.getRequestID(), this._onQueue);
        console.error("Invalid response object from main request code.");
      }
    } catch (e) { /*EXECUTION FAIL*/
      console.error('Exception when executing main request code.');
      console.exception(e);
      if (executionFailed) {
        this._queueResp = IKBadRequestResponseWithRollback(IKGlobals.ErrorResponseUnhandledError, IKGlobals.ErrorCode_APIError);
        this._queueResp.build(this._context, this.request.getRequestID(), this._onQueue);
      }
    } return executionFailed;
  }
  async _executeDBTransactions(safeExecution) {
    try {
      //start DB
      await this.db.connect();
      //start transaction
      await this.db.beginTransaction();
      //
      let executionFailed = await safeExecution();
      //Commit if not failed, rollback if failed
      if (!executionFailed) await this.db.commit();
      else {
        console.log("Rolling back DB transactions. Main code failed!");
        await this.db.rollback();
      }
      //Cleanup DB after execution
      await this.db.cleanup();
    } catch (e) {
      console.error('Exception when executing DB transactions.');
      console.log(e);
    }
  }
  async _executeLoggerFlush(safeExecution) {
    try {
      await safeExecution();
      await this.logger.flushLogs();
    } catch (e) {
      console.error('Exception when flushing logs.');
      console.exception(e);
    } finally {
      console.debug("Transaction ended");
    }
  }
  /* DB drivers support */
  _getDBDriver() {
    if (this._config.database) {
      if (this._config.database.type == IKGlobals.DBDrivers.SQL) return new IKDBSQL(this._config.database);
      else if (Object.keys(this._config.database).length > 0) return new IKDBDDB(this._config.database);
    } return null;
  }
}
