import AWS from "aws-sdk";
const abind = require('abind');
const stackTrace = require('stack-trace');
const stringify = require('json-stringify-safe');
//
import Utils from "./../API/IKUtils";
//
export const LOG_LEVELS = { DEBUG: 0, INFO: 1, WARN: 2, ERROR: 3 };
const LOG_STRINGS = ['DEBUG', 'INFO', 'WARN', 'ERROR'];
const PURE_CONSOLE = (console.flushLogs ? console.origin : console);
const DEFAULT_LOG_FUNCTION = PURE_CONSOLE.log.bind(PURE_CONSOLE);
//
export default class IKLogger {
  constructor(_config, _LOG_LEVEL, transactionID) {
    abind(this);
    //
    _LOG_LEVEL = LOG_LEVELS.DEBUG;
    if (!_LOG_LEVEL) _LOG_LEVEL = LOG_LEVELS.DEBUG;
    else {
      _LOG_LEVEL = LOG_STRINGS.indexOf(_LOG_LEVEL);
      if (_LOG_LEVEL == -1) _LOG_LEVEL = LOG_LEVELS.DEBUG;
    }
    //
    this.origin = PURE_CONSOLE;
    this._LOG_LEVEL = _LOG_LEVEL;
    this._s3Config = _config;
    this._transactionID = transactionID;
    this._logs = [];
    this._setupBindings();
    //
    console.log("Using logger with level: " + LOG_STRINGS[this._LOG_LEVEL]);
  }

  //Public
  debug() { this._processLog(LOG_LEVELS.DEBUG, arguments); }
  log() { this._processLog(LOG_LEVELS.INFO, arguments); }
  info() { this._processLog(LOG_LEVELS.INFO, arguments); }
  warning() { this._processLog(LOG_LEVELS.WARN, arguments); }
  warn() { this._processLog(LOG_LEVELS.WARN, arguments); }
  error() { this._processLog(LOG_LEVELS.ERROR, arguments); }
  exception(exception) {
    //format message
    let msg = [];
    //push exeception
    msg.push(exception + " -");
    //get args
    for (let arg of arguments) if (arg != exception) msg.push(arg);
    msg.push(exception.stack); //push exeception stack at the end
    //push into logs stack
    this._pushLog(LOG_LEVELS.ERROR, this._formattedLog(LOG_LEVELS.ERROR, msg, this._callerName(3)));
  }
  //
  cleanUp() { this._logs = []; }
  async flushLogs() {}


  //initialization
  _setupBindings() {
    global.console = this;
  }


  //Helpers
  _timestamp() {
    let d = new Date();
    return [Utils.toDoubleDigit(d.getMonth()+1), Utils.toDoubleDigit(d.getDate()), d.getFullYear()].join('/') +' ' +
           [Utils.toDoubleDigit(d.getHours()), Utils.toDoubleDigit(d.getMinutes()), Utils.toDoubleDigit(d.getSeconds())].join(':');
  }
  _formattedLog(level, msg, caller) {
    if (Utils.isOffline()) {
      return `  [${this._timestamp()} - ${LOG_STRINGS[level]}] [${caller}] ${msg.join(" ")}`;
    } else if (Utils.isIKAPIGateway() && this._transactionID) {
      return ` ${this._transactionID} [${LOG_STRINGS[level]}] [${caller}] ${msg.join(" ")}`;
    } else {
      return ` [${LOG_STRINGS[level]}] [${caller}] ${msg.join(" ")}`;
    }
  }
  _callerName(index) {
    let safeIndex = Math.min(index, stackTrace.get().length);
    if (stackTrace.get()[safeIndex]) {
      let callerName = (stackTrace.get()[safeIndex] ? stackTrace.get()[safeIndex].getTypeName() : null);
      if (!callerName) {
        callerName = stackTrace.get()[safeIndex].getFileName().split("/");
        callerName = callerName.slice(callerName.indexOf("src")).join("/");
      } return callerName + ":" + stackTrace.get()[safeIndex].getLineNumber();
    } return '';
  }
  _processLog(level, args) {
    if (level < this._LOG_LEVEL) return;
    //get args
    let msg = [];
    for (let arg of args) msg.push(((typeof arg === 'object' && !(arg instanceof Error)) ? stringify(arg, null, 2) : arg));
    //push into logs stack
    // todo: improve error stack
    this._pushLog(level, this._formattedLog(level, msg, this._callerName(3)));
  }
  _pushLog(level, fMsg) {
    //push into logs stack
    this._logs.push(fMsg);
    DEFAULT_LOG_FUNCTION.apply(PURE_CONSOLE, [fMsg]);
  }
}
